'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _toArray2 = require('babel-runtime/helpers/toArray');

var _toArray3 = _interopRequireDefault(_toArray2);

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

exports.neo4jgraphql = neo4jgraphql;
exports.cypherQuery = cypherQuery;

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var _ = require('lodash');


var returnTypeEnum = {
  OBJECT: 0,
  ARRAY: 1
};

function neo4jgraphql(object, params, context, resolveInfo) {

  // const returnTypeEnum = {
  //   OBJECT: 0,
  //   ARRAY: 1
  // };

  var type = innerType(resolveInfo.returnType).toString(),
      variable = type.charAt(0).toLowerCase() + type.slice(1);

  var query = cypherQuery(params, context, resolveInfo);

  console.log(query);

  var returnType = resolveInfo.returnType.toString().startsWith("[") ? returnTypeEnum.ARRAY : returnTypeEnum.OBJECT;

  var session = context.driver.session();
  var data = session.run(query, params).then(function (result) {

    if (returnType === returnTypeEnum.ARRAY) {
      return result.records.map(function (record) {
        return record.get(variable);
      });
    } else if (returnType === returnTypeEnum.OBJECT) {
      if (result.records.length > 0) {
        // FIXME: use one of the new neo4j-driver consumers when upgrading neo4j-driver package
        return result.records[0].get(variable);
      } else {
        return null;
      }
    }
  });

  return data;
};

function cypherQuery(params, context, resolveInfo) {
  var pageParams = {
    "first": params['first'] === undefined ? -1 : params['first'],
    "offset": params['offset'] || 0
  };

  delete params['first'];
  delete params['offset'];

  var type = innerType(resolveInfo.returnType).toString(),
      variable = type.charAt(0).toLowerCase() + type.slice(1),
      schemaType = resolveInfo.schema.getType(type);

  var filteredFieldNodes = _.filter(resolveInfo.fieldNodes, function (o) {
    if (o.name.value === resolveInfo.fieldName) {
      return true;
    }
  });

  // FIXME: how to handle multiple fieldNode matches
  var selections = filteredFieldNodes[0].selectionSet.selections;

  var wherePredicate = '';
  if (_.has(params, '_id')) {
    wherePredicate = 'WHERE ID(' + variable + ')=' + params._id + ' ';
    delete params._id;
  }

  var argString = (0, _stringify2.default)(params).replace(/\"([^(\")"]+)\":/g, "$1:"); // FIXME: support IN for multiple values -> WHERE
  var query = 'MATCH (' + variable + ':' + type + ' ' + argString + ') ' + wherePredicate;

  query = query + ('RETURN ' + variable + ' {') + buildCypherSelection('', selections, variable, schemaType, resolveInfo); // ${variable} { ${selection} } as ${variable}`;

  query = query + ('} AS ' + variable);

  query = query + (' SKIP ' + pageParams.offset + (pageParams.first > -1 ? " LIMIT " + pageParams.first : ""));

  return query;
}

function buildCypherSelection(initial, selections, variable, schemaType, resolveInfo) {
  //FIXME: resolveInfo not needed

  if (selections.length === 0) {
    return initial;
  }

  var _selections = (0, _toArray3.default)(selections),
      headSelection = _selections[0],
      tailSelections = _selections.slice(1);

  var fieldName = headSelection.name.value;
  if (!schemaType.getFields()[fieldName]) {
    // meta field type
    return buildCypherSelection(tailSelections.length === 0 ? initial.substring(initial.lastIndexOf(','), 0) : initial, tailSelections, variable, schemaType, resolveInfo);
  }
  var fieldType = schemaType.getFields()[fieldName].type;

  var inner = innerType(fieldType); // for target "type" aka label

  var fieldHasCypherDirective = schemaType.getFields()[fieldName].astNode.directives.filter(function (e) {
    return e.name.value === 'cypher';
  }).length > 0;

  if (fieldHasCypherDirective) {

    var statement = schemaType.getFields()[fieldName].astNode.directives.find(function (e) {
      return e.name.value === 'cypher';
    }).arguments.find(function (e) {
      return e.name.value === 'statement';
    }).value.value;

    if (inner.constructor.name === "GraphQLScalarType") {

      return buildCypherSelection(initial + (fieldName + ': apoc.cypher.runFirstColumn("' + statement + '", ' + (0, _utils.cypherDirectiveArgs)(variable, headSelection, schemaType) + ', false)' + (tailSelections.length > 0 ? ',' : '')), tailSelections, variable, schemaType, resolveInfo);
    } else {
      // similar: [ x IN apoc.cypher.runFirstColumn("WITH {this} AS this MATCH (this)--(:Genre)--(o:Movie) RETURN o", {this: movie}, true) |x {.title}][1..2])

      var nestedVariable = variable + '_' + fieldName;
      var skipLimit = computeSkipLimit(headSelection);
      var fieldIsList = !!fieldType.ofType;

      return buildCypherSelection(initial + (fieldName + ': ' + (fieldIsList ? "" : "head(") + '[ x IN apoc.cypher.runFirstColumn("' + statement + '", ' + (0, _utils.cypherDirectiveArgs)(variable, headSelection, schemaType) + ', true) | x {' + buildCypherSelection('', headSelection.selectionSet.selections, nestedVariable, inner, resolveInfo) + '}]' + (fieldIsList ? "" : ")") + skipLimit + ' ' + (tailSelections.length > 0 ? ',' : '')), tailSelections, variable, schemaType, resolveInfo);
    }
  } else if (innerType(fieldType).constructor.name === "GraphQLScalarType") {
    return buildCypherSelection(initial + (' .' + fieldName + ' ' + (tailSelections.length > 0 ? ',' : '')), tailSelections, variable, schemaType, resolveInfo);
  } else {
    // field is an object
    var _nestedVariable = variable + '_' + fieldName,
        _skipLimit = computeSkipLimit(headSelection),
        relationDirective = schemaType.getFields()[fieldName].astNode.directives.find(function (e) {
      return e.name.value === 'relation';
    }),
        relType = relationDirective.arguments.find(function (e) {
      return e.name.value === 'name';
    }).value.value,
        relDirection = relationDirective.arguments.find(function (e) {
      return e.name.value === 'direction';
    }).value.value;

    var returnType = fieldType.toString().startsWith("[") ? returnTypeEnum.ARRAY : returnTypeEnum.OBJECT;

    return buildCypherSelection(initial + (fieldName + ': ' + (returnType === returnTypeEnum.OBJECT ? 'head(' : '') + '[(' + variable + ')' + (relDirection === 'in' || relDirection === 'IN' ? '<' : '') + '-[:' + relType + ']-' + (relDirection === 'out' || relDirection === 'OUT' ? '>' : '') + '(' + _nestedVariable + ':' + inner.name + ') | ' + _nestedVariable + ' {' + buildCypherSelection('', headSelection.selectionSet.selections, _nestedVariable, inner, resolveInfo) + '}]' + (returnType === returnTypeEnum.OBJECT ? ')' : '') + _skipLimit + ' ' + (tailSelections.length > 0 ? ',' : '')), tailSelections, variable, schemaType, resolveInfo);
  }
}

function innerType(type) {
  return type.ofType ? innerType(type.ofType) : type;
}

function argumentValue(selection, name) {
  var arg = selection.arguments.find(function (a) {
    return a.name.value === name;
  });
  return arg === undefined ? null : arg.value.value;
}

function computeSkipLimit(selection) {
  var first = argumentValue(selection, "first");
  var offset = argumentValue(selection, "offset");

  if (first === null && offset === null) return "";
  if (offset === null) return '[..' + first + ']';
  if (first === null) return '[' + offset + '..]';
  return '[' + offset + '..' + (parseInt(offset) + parseInt(first)) + ']';
}